#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""Tests for the request module."""

# Method could be a function
# pylint: disable-msg=R0201

# Too many public methods
# pylint: disable-msg=R0904

# Standard modules
try:
  import cStringIO as stringIO
except ImportError:
  import io as stringIO

import unittest
import urllib

# Unittest target
from uweb3 import request


class IndexedFieldStorageTest(unittest.TestCase):
  """Comprehensive testing of the IndexedFieldStorage object."""

  def CreateFieldStorage(self, data):
    """Returns an IndexedFieldStorage object constructed from the given data."""
    return request.IndexedFieldStorage(stringIO.StringIO(data),
                                       environ={'REQUEST_METHOD': 'POST'})

  def testEmptyStorage(self):
    """An empty IndexedFieldStorage is empty and is boolean False"""
    ifs = self.CreateFieldStorage('')
    self.assertFalse(ifs)

  def testBasicStorage(self):
    """A basic IndexedFieldStorage has the proper key + value pair"""
    ifs = self.CreateFieldStorage('key=value')
    self.assertEqual(ifs.getfirst('key'), 'value')
    self.assertEqual(ifs.getlist('key'), ['value'])

  def testMissingKey(self):
    """getfirst / getlist for missing keys return proper defaults"""
    ifs = self.CreateFieldStorage('')
    self.assertEqual(ifs.getfirst('missing'), None)
    self.assertEqual(ifs.getfirst('missing', 'signal'), 'signal')
    self.assertEqual(ifs.getlist('missing'), [])
    # getlist has no default argument

  def testMultipleKeyOrder(self):
    """getlist returns keys in the order they were given to the FieldStorage"""
    args = ['1', '3', '2']
    ifs = self.CreateFieldStorage('arg=1&arg=3&arg=2')
    self.assertEqual(ifs.getlist('arg'), args)

  def testUrlEncoding(self):
    """IndexedFieldStorage can handle utf-8 encoded forms"""
    string = u'We ♥ Unicode'
    try:
      data = urllib.urlencode({'q': string})
    except AttributeError:
      data = urllib.parse.urlencode({'q': string})
    ifs = self.CreateFieldStorage(data)
    self.assertEqual(ifs.getfirst('q'), string)

  def testDictionaryFunctionality(self):
    """The indexing of IndexedFieldStorage allows dict-like assignment"""
    data = 'd[name]=Arthur&d[type]=King'
    ifs = self.CreateFieldStorage(data)
    self.assertEqual(ifs.getfirst('d'), {'name': 'Arthur', 'type': 'King'})

  def testDictionaryAndMultipleValues(self):
    """Dictionaries and regular values can be combined; dict is item no. #1"""
    data = 'd=third&d[first]=1&d[second]=2&d=fourth'
    ifs = self.CreateFieldStorage(data)
    form_data = ifs.getlist('d')
    self.assertEqual(form_data[0], {'first': '1', 'second': '2'})
    self.assertEqual(form_data[1], 'third')
    self.assertEqual(form_data[2], 'fourth')


if __name__ == '__main__':
  unittest.main(testRunner=unittest.TextTestRunner(verbosity=2))



#!/usr/bin/python3
"""This file contains the Base connector for model connections and imports all
available connectors."""

__author__ = 'Jan Klopper (jan@underdark.nl)'
__version__ = 0.1

class Connector(object):
  """Base Connector class, subclass from this to create your own connectors.
  Usually the name of your class is used to lookup its config in the
  configuration file, or the database or local filename.

  Connectors based on this class are Usually Singletons. One global connection
  is kept alive, and multiple model classes use it to connect to their
  respective tables, cookies, or files.
  """
  _NAME = None

  @classmethod
  def Name(cls):
    """Returns the 'connector' name, which is usally used to lookup its config
     in the config file.

    If this is not explicitly defined by the class constant `_TABLE`, the return
    value will be the class name with the first letter lowercased.
    """
    if cls._NAME:
      return cls._NAME
    name = cls.__name__
    return name[0].lower() + name[1:]

  def Disconnect(self):
    """Standard interface to disconnect from data source"""
    raise NotImplementedError

  def Rollback(self):
    """Standard interface to rollback any pending commits"""
    raise NotImplementedError

from .SignedCookie import SignedCookie
from .Mysql import Mysql
from .Mongo import Mongo
from .Sqlite import Sqlite
from .SqlAlchemy import SqlAlchemy



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""µWeb3 request module."""

# Standard modules
import cgi
import sys
import urllib
import io
from urllib.parse import parse_qs, parse_qsl
import io as stringIO
import http.cookies as cookie
import re
import json

# uWeb modules
from . import response


class CookieTooBigError(Exception):
  """Error class for cookie when size is bigger than 4096 bytes"""


class Cookie(cookie.SimpleCookie):
  """Cookie class that uses the most specific value for a cookie name.

  According to RFC2965 (http://tools.ietf.org/html/rfc2965):
      If multiple cookies satisfy the criteria above, they are ordered in
      the Cookie header such that those with more specific Path attributes
      precede those with less specific.  Ordering with respect to other
      attributes (e.g., Domain) is unspecified.

  This class adds this behaviour to cookie parsing. That is, a key:value pair
  WILL NOT overwrite an already existing (and thus more specific) pair.

  N.B.: this class assumes the given cookie to follow the standards outlined in
  the RFC. At the moment (2011Q1) this assumption proves to be correct for both
  Chromium (and likely Webkit in general) and Firefox. Other browsers have not
  been tested, and might possibly deviate from the suggested standard.
  As such, it's recommended not to re-use the cookie name with different values
  for different paths.
  """
  # Unfortunately this works by redefining a private method.
  def _BaseCookie__set(self, key, real_value, coded_value):
    """Inserts a morsel into the Cookie, strictly on the first occurrance."""
    if key not in self:
      morsel = cookie.Morsel()
      morsel.set(key, real_value, coded_value)
      dict.__setitem__(self, key, morsel)


class Request:
  def __init__(self, env, registry):
    self.env = env
    self.headers = dict(self.headers_from_env(env))
    self.registry = registry
    self._out_headers = []
    self._out_status = 200
    self._response = None
    self.method = self.env['REQUEST_METHOD']
    self.vars = {
        'cookie': {
            name: value.value
            for name, value in Cookie(self.env.get('HTTP_COOKIE')).items()
        },
        'get': QueryArgsDict(parse_qs(self.env['QUERY_STRING'])),
    }
    self.env['host'] = self.headers.get('Host', '')
    if self.method in ('POST', 'PUT', 'DELETE'):
      request_body_size = 0
      try:
        request_body_size = int(self.env.get('CONTENT_LENGTH', 0))
      except Exception:
        pass
      request_payload = self.env['wsgi.input'].read(request_body_size)
      self.input = request_payload
      if self.env.get('CONTENT_TYPE', '') == 'application/json':
        self.vars[self.method.lower()] = json.loads(request_payload)
      else:
        self.vars[self.method.lower()] = IndexedFieldStorage(stringIO.StringIO(request_payload.decode("utf-8")),
             environ={'REQUEST_METHOD': 'POST'})

  @property
  def path(self):
    return self.env['PATH_INFO']

  @property
  def response(self):
    if self._response is None:
      self._response = response.Response(headers=self._out_headers)
    return self._response

  def Redirect(self, location, httpcode=307):
    REDIRECT_PAGE = ('<!DOCTYPE html><html><head><title>Page moved</title></head>'
                     '<body>Page moved, please follow <a href="{}">this link</a>'
                     '</body></html>').format(location)

    headers = {'Location': location}
    if self.response.headers.get('Set-Cookie'):
      headers['Set-Cookie'] = self.response.headers.get('Set-Cookie')
    return response.Response(
      content=REDIRECT_PAGE,
      content_type=self.response.headers.get('Content-Type', 'text/html'),
      httpcode=httpcode,
      headers=headers
      )

  def headers_from_env(self, env):
    for key, value in env.items():
      if key.startswith('HTTP_'):
        yield key[5:].lower().replace('_', '-'), value

  def AddCookie(self, key, value, **attrs):
    """Adds a new cookie header to the response.

    Arguments:
      @ key: str
        The name of the cookie.
      @ value: str
        The actual value to store in the cookie.
      % expires: str ~~ None
        The date + time when the cookie should expire. The format should be:
        "Wdy, DD-Mon-YYYY HH:MM:SS GMT" and the time specified in UTC.
        The default means the cookie never expires.
        N.B. Specifying both this and `max_age` leads to undefined behavior.
      % path: str ~~ '/'
        The path for which this cookie is valid. This default ('/') is different
        from the rule stated on Wikipedia: "If not specified, they default to
        the domain and path of the object that was requested".
      % domain: str ~~ None
        The domain for which the cookie is valid. The default is that of the
        requested domain.
      % max_age: int
        The number of seconds this cookie should be used for. After this period,
        the cookie should be deleted by the client.
        N.B. Specifying both this and `expires` leads to undefined behavior.
      % secure: boolean
        When True, the cookie is only used on https connections.
      % httponly: boolean
        When True, the cookie is only used for http(s) requests, and is not
        accessible through Javascript (DOM).
    """
    if isinstance(value, (str)) and len(value.encode('utf-8')) >= 4096:
      raise CookieTooBigError("Cookie is larger than 4096 bytes and wont be set")

    new_cookie = Cookie({key: value})
    if 'max_age' in attrs:
      attrs['max-age'] = attrs.pop('max_age')
    new_cookie[key].update(attrs)
    self.AddHeader('Set-Cookie', new_cookie[key].OutputString())

  def AddHeader(self, name, value):
    if name == 'Set-Cookie':
      if not self.response.headers.get('Set-Cookie'):
        self.response.headers['Set-Cookie'] = [value]
        return
      self.response.headers['Set-Cookie'].append(value)
      return
    self.response.AddHeader(name, value)

  def DeleteCookie(self, name):
    """Deletes cookie by name

    Arguments
    @ name: str
    """
    self.AddHeader('Set-Cookie', '{}=deleted; expires=Thu, 01 Jan 1970 00:00:00 GMT;'.format(name))


class IndexedFieldStorage(cgi.FieldStorage):
  """Adaption of cgi.FieldStorage with a few specific changes.

  Notable differences with cgi.FieldStorage:
    1) `environ.QUERY_STRING` does not add to the returned FieldStorage
       This way we maintain a strict separation between POST and GET variables.
    2) Field names in the form 'foo[bar]=baz' will generate a dictionary:
         foo = {'bar': 'baz'}
       Multiple statements of the form 'foo[%s]' will expand this dictionary.
       Multiple occurrances of 'foo[bar]' will result in unspecified behavior.
    3) Automatically attempts to parse all input as UTF8. This is the proposed
       standard as of 2005: http://tools.ietf.org/html/rfc3986.
  """
  FIELD_AS_ARRAY = re.compile(r'(.*)\[(.*)\]')
  def iteritems(self):
    return ((key, self.getlist(key)) for key in self)

  def items(self):
    return list(self.iteritems())

  def read_urlencoded(self):
    indexed = {}
    self.list = []
    for field, value in parse_qsl(self.fp.read(self.length),
                                      self.keep_blank_values,
                                      self.strict_parsing):
      if self.FIELD_AS_ARRAY.match(str(field)):
        field_group, field_key = self.FIELD_AS_ARRAY.match(field).groups()
        indexed.setdefault(field_group, cgi.MiniFieldStorage(field_group, {}))
        indexed[field_group].value[field_key] = value
      else:
        self.list.append(cgi.MiniFieldStorage(field, value))
    self.list = list(indexed.values()) + self.list
    self.skip_lines()

  def __repr__(self):
    return "{%s}" % ','.join("'%s': '%s'" % (k, v if len(v) > 1 else v[0]) for k, v in self.iteritems())

  @property
  def __dict__(self):
    return {
        key: value if len(value) > 1 else value[0]
        for key, value in self.iteritems()
    }


class QueryArgsDict(dict):
  def getfirst(self, key, default=None):
    """Returns the first value for the requested key, or a fallback value."""
    try:
      return self[key][0]
    except KeyError:
      return default

  def getlist(self, key):
    """Returns a list with all values that were given for the requested key.

    N.B. If the given key does not exist, an empty list is returned.
    """
    try:
      return self[key]
    except KeyError:
      return []


def return_real_remote_addr(env):
  """Returns the remote ip-address,
  if there is a proxy involved it will take the last IP addres from the HTTP_X_FORWARDED_FOR list
  """
  try:
    return env['HTTP_X_FORWARDED_FOR'].split(',')[-1].strip()
  except KeyError:
    return env['REMOTE_ADDR']



from itertools import chain

from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import reconstructor, sessionmaker
from sqlalchemy.orm.session import object_session

from uweb3.model import NotExistError


class AlchemyBaseRecord:
  def __init__(self, session, record):
    self.session = session
    self._BuildClassFromRecord(record)

  def _BuildClassFromRecord(self, record):
    if not isinstance(record, dict):
      return
    for key, value in record.items():
      if key not in self.__table__.columns.keys():
        raise AttributeError(f"Key '{key}' not specified in class '{self.__class__.__name__}'")
      setattr(self, key, value)
    if self.session:
      try:
        self.session.add(self)
      except:
        self.session.rollback()
        raise
      else:
        self.session.commit()

  def __hash__(self):
    """Returns the hashed value of the key."""
    return hash(self.key)

  def __del__(self):
    """Cleans up the connection at the end of its life cycle"""
    self.session.close()

  def __repr__(self):
    s = {}
    for key in self.__table__.columns.keys():
      value = getattr(self, key)
      if value:
        s[key] = value
    return f'{type(self).__name__}({s})'

  def __eq__(self, other):
    if type(self) != type(other):
      return False  # Types must be the same.
    elif not (self.key == other.key is not None):
      return False  # Records should have the same non-None primary key value.
    elif len(self) != len(other):
      return False  # Records must contain the same number of objects.
    for key in self.__table__.columns.keys():
      value = getattr(self, key)
      other_value = getattr(other, key)
      if isinstance(self, AlchemyBaseRecord) != isinstance(other, AlchemyBaseRecord):
        # Only one of the two is a BaseRecord instance
        if (isinstance(self, AlchemyBaseRecord) and value.key != other_value or
            isinstance(other, AlchemyBaseRecord) and other_value.key != value):
          return False
      elif value != other_value:
        return False
    return True

  def __ne__(self, other):
    """Returns the proper inverse of __eq__."""
    # Without this, the non-equal checks used in __eq__ will not work,
    # and the  `!=` operator would not be the logical inverse of `==`.
    return not self == other

  def __len__(self):
    return len({
        col: getattr(self, col)
        for col in self.__table__.columns.keys() if getattr(self, col)
    })

  def __int__(self):
    """Returns the integer key value of the Record.

    For record objects where the primary key value is not (always) an integer,
    this function will raise an error in the situations where it is not.
    """
    key_val = self.key
    if not isinstance(key_val, (int)):
      # We should not truncate floating point numbers.
      # Nor turn strings of numbers into an integer.
      raise ValueError('The primary key is not an integral number.')
    return key_val

  def copy(self):
    """Returns a shallow copy of the Record that is a new functional Record."""
    import copy
    return copy.copy(self)

  def deepcopy(self):
    import copy
    return copy.deepcopy(self)

  def __gt__(self, other):
    """Index of this record is greater than the other record's.

    This requires both records to be of the same record class.
    """
    if type(self) == type(other):
      return self.key > other.key
    return NotImplemented

  def __ge__(self, other):
    """Index of this record is greater than, or equal to, the other record's.

    This requires both records to be of the same record class.
    """
    if type(self) == type(other):
      return self.key >= other.key
    return NotImplemented

  def __lt__(self, other):
    """Index of this record is smaller than the other record's.

    This requires both records to be of the same record class.
    """
    if type(self) == type(other):
      return self.key < other.key
    return NotImplemented

  def __le__(self, other):
    """Index of this record is smaller than, or equal to, the other record's.

    This requires both records to be of the same record class.
    """
    if type(self) == type(other):
      return self.key <= other.key
    return NotImplemented

  def __getitem__(self, field):
    return getattr(self, field)

  def iteritems(self):
    """Yields all field+value pairs in the Record.

    This automatically loads in relationships.
    """
    return chain(((key, getattr(self, key)) for key in self.__table__.columns.keys()),
    ((child[0], getattr(self, child[0])) for child in inspect(type(self)).relationships.items()))

  def itervalues(self):
    """Yields all values in the Record, loads relationships"""
    return chain((getattr(self, key) for key in self.__table__.columns.keys()),
                 (getattr(self, child[0]) for child in inspect(type(self)).relationships.items()))

  def items(self):
    """Returns a list of field+value pairs in the Record.

    This automatically loads in relationships.
    """
    return list(self.iteritems())

  def values(self):
    """Returns a list of values in the Record, loading foreign references."""
    return list(self.itervalues())

  @property
  def key(self):
    return getattr(self, inspect(type(self)).primary_key[0].name)

  @classmethod
  def TableName(cls):
    """Returns the database table name for the Record class."""
    return cls.__tablename__

  @classmethod
  def _AlchemyRecordToDict(cls, record):
    """Turns the values of a given class into a dictionary. Doesn't trigger
    automatic loading of child classes.

    Arguments:
      @ record: cls
        AlchemyBaseRecord class that is retrieved from a database query
    Returns
      dict: dictionary with all table columns and values
      None: when record is empty
    """
    if not isinstance(record, type(None)):
      return {col: getattr(record, col) for col in record.__table__.columns.keys()}
    return None

  @reconstructor
  def reconstruct(self):
    """This is called instead of __init__ when the result comes from the database"""
    self.session = object_session(self)

  @classmethod
  def _PrimaryKeyCondition(cls, target):
    """Returns the name of the primary key of given class

    Arguments:
      @ target: cls
        Class that you want to know the primary key name from
    """
    return getattr(cls, inspect(cls).primary_key[0].name)

class AlchemyRecord(AlchemyBaseRecord):
  """ """
  @classmethod
  def FromPrimary(cls, session, p_key):
    """Finds record based on given class and supplied primary key.

    Arguments:
      @ session: sqlalchemy session object
        Available in the pagemaker with self.session
      @ p_key: integer
        primary_key of the object to delete
    Returns
      cls
      None
    """
    try:
      record = session.query(cls).filter(cls._PrimaryKeyCondition(cls) == p_key).first()
    except:
      session.rollback()
      raise
    else:
      if not record:
        raise NotExistError(f"Record with primary key {p_key} does not exist")
      return record

  @classmethod
  def DeletePrimary(cls, session, p_key):
    """Deletes record base on primary key from given class.

    Arguments:
      @ session: sqlalchemy session object
        Available in the pagemaker with self.session
      @ p_key: integer
        primary_key of the object to delete

    Returns:
      isdeleted: boolean
    """
    try:
      isdeleted = session.query(cls).filter(cls._PrimaryKeyCondition(cls) == p_key).delete()
    except:
      session.rollback()
      raise
    else:
      session.commit()
      return isdeleted

  @classmethod
  def Create(cls, session, record):
    """Creates a new instance and commits it to the database

    Arguments:
      @ session: sqlalchemy session object
        Available in the pagemaker with self.session
      @ record: dict
        Dictionary with all key:value pairs that are required for the db record
    Returns:
      cls
    """
    return cls(session, record)

  @classmethod
  def List(cls, session, conditions=None, limit=None, offset=None,
           order=None, yield_unlimited_total_first=False):
    """Yields a Record object for every table entry.

    Arguments:
      @ session: sqlalchemy session object
        Available in the pagemaker with self.session
      % conditions: list
        Optional query portion that will be used to limit the list of results.
        If multiple conditions are provided, they are joined on an 'AND' string.
        For example: conditions=[User.id <= 10, User.id >=]
      % limit: int ~~ None
        Specifies a maximum number of items to be yielded. The limit happens on
        the database side, limiting the query results.
      % offset: int ~~ None
        Specifies the offset at which the yielded items should start. Combined
        with limit this enables proper pagination.
      % order: tuple of operants
        For example the User class has 3 fields; id, username, password. We can pass
        the field we want to order on to the tuple like so;
        (User.id.asc(), User.username.desc())
      % yield_unlimited_total_first: bool ~~ False
        Instead of yielding only Record objects, the first item returned is the
        number of results from the query if it had been executed without limit.

    Returns:
      integer: integer with length of results.
      list: List of classes from request type
    """
    try:
      query = session.query(cls)
      if conditions:
        for condition in conditions:
          query = query.filter(condition)
      if order:
        for item in order:
          query = query.order_by(item)
      if limit:
        query = query.limit(limit)
      if offset:
        query = query.offset(offset)
      result = query.all()
    except:
      session.rollback()
      raise
    else:
      if yield_unlimited_total_first:
        return len(result)
      return result

  @classmethod
  def Update(cls, session, conditions, values):
    """Update table based on conditions.

    Arguments:
      @ session: sqlalchemy session object
          Available in the pagemaker with self.session
      @ conditions: list|tuple
        for example: [User.id > 2, User.id < 100]
      @ values: dict
        for example: {User.username: 'value'}
    """
    try:
      query = session.query(cls)
      for condition in conditions:
        query = query.filter(condition)
      query = query.update(values)
    except:
      session.rollback()
      raise
    else:
      session.commit()

  def Delete(self):
    """Delete current instance from the database"""
    try:
      isdeleted = self.session.query(type(self)).filter(self._PrimaryKeyCondition(self) == self.key).delete()
    except:
      self.session.rollback()
      raise
    else:
      self.session.commit()
      return isdeleted

  def Save(self):
    """Saves any changes made in the current record. Sqlalchemy automatically detects
    these changes and only updates the changed values. If no values are present
    no query will be commited."""
    self.session.commit()


class AlchemyRecord(AlchemyBaseRecord):
  """ """
  @classmethod
  def FromPrimary(cls, session, p_key):
    """Finds record based on given class and supplied primary key.

    Arguments:
      @ session: sqlalchemy session object
        Available in the pagemaker with self.session
      @ p_key: integer
        primary_key of the object to delete
    Returns
      cls
      None
    """
    try:
      record = session.query(cls).filter(cls._PrimaryKeyCondition(cls) == p_key).first()
    except:
      session.rollback()
      raise
    else:
      if not record:
        raise NotExistError(f"Record with primary key {p_key} does not exist")
      return record

  @classmethod
  def DeletePrimary(cls, session, p_key):
    """Deletes record base on primary key from given class.

    Arguments:
      @ session: sqlalchemy session object
        Available in the pagemaker with self.session
      @ p_key: integer
        primary_key of the object to delete

    Returns:
      isdeleted: boolean
    """
    try:
      isdeleted = session.query(cls).filter(cls._PrimaryKeyCondition(cls) == p_key).delete()
    except:
      session.rollback()
      raise
    else:
      session.commit()
      return isdeleted

  @classmethod
  def Create(cls, session, record):
    """Creates a new instance and commits it to the database

    Arguments:
      @ session: sqlalchemy session object
        Available in the pagemaker with self.session
      @ record: dict
        Dictionary with all key:value pairs that are required for the db record
    Returns:
      cls
    """
    return cls(session, record)

  @classmethod
  def List(cls, session, conditions=None, limit=None, offset=None,
           order=None, yield_unlimited_total_first=False):
    """Yields a Record object for every table entry.

    Arguments:
      @ session: sqlalchemy session object
        Available in the pagemaker with self.session
      % conditions: list
        Optional query portion that will be used to limit the list of results.
        If multiple conditions are provided, they are joined on an 'AND' string.
        For example: conditions=[User.id <= 10, User.id >=]
      % limit: int ~~ None
        Specifies a maximum number of items to be yielded. The limit happens on
        the database side, limiting the query results.
      % offset: int ~~ None
        Specifies the offset at which the yielded items should start. Combined
        with limit this enables proper pagination.
      % order: tuple of operants
        For example the User class has 3 fields; id, username, password. We can pass
        the field we want to order on to the tuple like so;
        (User.id.asc(), User.username.desc())
      % yield_unlimited_total_first: bool ~~ False
        Instead of yielding only Record objects, the first item returned is the
        number of results from the query if it had been executed without limit.

    Returns:
      integer: integer with length of results.
      list: List of classes from request type
    """
    try:
      query = session.query(cls)
      if conditions:
        for condition in conditions:
          query = query.filter(condition)
      if order:
        for item in order:
          query = query.order_by(item)
      if limit:
        query = query.limit(limit)
      if offset:
        query = query.offset(offset)
      result = query.all()
    except:
      session.rollback()
      raise
    else:
      if yield_unlimited_total_first:
        return len(result)
      return result

  @classmethod
  def Update(cls, session, conditions, values):
    """Update table based on conditions.

    Arguments:
      @ session: sqlalchemy session object
          Available in the pagemaker with self.session
      @ conditions: list|tuple
        for example: [User.id > 2, User.id < 100]
      @ values: dict
        for example: {User.username: 'value'}
    """
    try:
      query = session.query(cls)
      for condition in conditions:
        query = query.filter(condition)
      query = query.update(values)
    except:
      session.rollback()
      raise
    else:
      session.commit()

  def Delete(self):
    """Delete current instance from the database"""
    try:
      isdeleted = self.session.query(type(self)).filter(self._PrimaryKeyCondition(self) == self.key).delete()
    except:
      self.session.rollback()
      raise
    else:
      self.session.commit()
      return isdeleted

  def Save(self):
    """Saves any changes made in the current record. Sqlalchemy automatically detects
    these changes and only updates the changed values. If no values are present
    no query will be commited."""
    self.session.commit()



#!/usr/bin/python3
"""uWeb3 installer."""

import os
import re
from setuptools import setup, find_packages

REQUIREMENTS = [
    'PyMySQL',
    'pytz'
]

def Description():
  """Returns the contents of the README.md file as description information."""
  with open(os.path.join(os.path.dirname(__file__), 'README.md')) as r_file:
    return r_file.read()


def Version():
  """Returns the version of the library as read from the __init__.py file"""
  main_lib = os.path.join(os.path.dirname(__file__), 'uweb3', '__init__.py')
  with open(main_lib) as v_file:
    return re.match(".*__version__ = '(.*?)'", v_file.read(), re.S).group(1)


setup(
    name='uWebthree',
    version=Version(),
    description='uWeb, python3, uswgi compatible micro web platform',
    long_description=Description(),
    long_description_content_type='text/markdown',
    license='ISC',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    author='Jan Klopper',
    author_email='jan@underdark.nl',
    url='https://github.com/underdark.nl/uweb3',
    keywords='minimal python web framework',
    packages=find_packages(),
    include_package_data=True,
    install_requires=REQUIREMENTS,
    python_requires='>=3.5')



#!/usr/bin/python3
"""This file contains the connector for SqlAlchemy."""
__author__ = 'Jan Klopper (janunderdark.nl)'
__version__ = 0.1

from . import Connector

class SqlAlchemy(Connector):
  """Adds MysqlAlchemy connection to ConnectionManager."""

  def __init__(self,  config, options, request, debug=False):
    """Returns a Mysql database connection wrapped in a SQLAlchemy session."""
    from sqlalchemy.orm import sessionmaker
    self.debug = debug
    self.options = {'host': 'localhost',
                    'user': None,
                    'password': None,
                    'database': ''}
    try:
      self.options = options[self.Name()]
    except KeyError:
      pass
    Session = sessionmaker()
    Session.configure(bind=self.engine, expire_on_commit=False)
    try:
      self.connection = Session()
    except Exception as e:
      raise ConnectionError('Connection to "%s" of type "%s" resulted in: %r' % (self.Name(), type(self), e))

  def engine(self):
    from sqlalchemy import create_engine
    return create_engine('mysql://{username}:{password}@{host}/{database}'.format(
      username=self.options.get('user'),
      password=self.options.get('password'),
      host=self.options.get('host', 'localhost'),
      database=self.options.get('database')),
      pool_size=5,
      max_overflow=0,
      encoding=self.options.get('charset', 'utf8'),)



#!/usr/bin/python3
"""Testsuite for the SQL Result abstraction module."""
__author__ = 'Elmer de Looff <elmer@underdark.nl>'
__version__ = '0.6'

# Too many public methods
# pylint: disable-msg=R0904

# Statement seems to have no effect
# pylint: disable-msg=W0104

# Unused variable %
# pylint: disable-msg=W0612

# Access to a protected member % of a client class
# pylint: disable-msg=W0212

# Method name too long
# pylint: disable-msg=C0103

# Standard modules
import unittest

# Unittest target
from . import sqlresult


class ResultRowBasicOperation(unittest.TestCase):
  """Ensure basic functionality and usability."""
  def setUp(self):
    """Set up a persistent test environment."""
    names = ['name', 'nationality', 'age']
    values = ('Elmer', 'Dutch', 24)
    self.row = sqlresult.ResultRow(names, values)

  def testFieldErrorIsLookupError(self):
    """SqlResult's FieldError is a subclass of LookupError"""
    self.assertTrue(issubclass(sqlresult.FieldError, LookupError))

  def testRepresentation(self):
    """ResultRow has representation methods that work."""
    repr(self.row)
    str(self.row)

  def testFalseWhenEmpty(self):
    """ResultRow is boolean False when empty."""
    row = sqlresult.ResultRow([], [])
    self.assertFalse(row)

  def testTrueWhenFilled(self):
    """ResultRow is boolean True when it has contents."""
    self.assertTrue(self.row)

  def testLength(self):
    """ResultRow is aware of its own length."""
    self.assertEquals(len(self.row), 3)

  def testGetByIndex(self):
    """ResultRow can retrieve data based on an index."""
    self.assertEquals(self.row[0], 'Elmer')
    self.assertEquals(self.row[2], 24)

  def testGetByKey(self):
    """ResultRow can retrieve data based on a key."""
    self.assertEquals(self.row['nationality'], 'Dutch')
    self.assertEquals(self.row['age'], 24)

  def testGetIndexError(self):
    """ResultRow throws FieldError when a bad index or bad key is requested."""
    self.assertRaises(sqlresult.FieldError, self.row.__getitem__, 4)
    self.assertRaises(sqlresult.FieldError, self.row.__getitem__, 'badkey')

  def testGetMethodIndex(self):
    """ResultRow's Get method works properly for indices."""
    self.assertEquals(self.row.get(1), 'Dutch')
    self.assertEquals(self.row.get(4), None)

  def testGetMethodKey(self):
    """ResultRow's Get method works properly for keys."""
    self.assertEquals(self.row.get('age'), 24)
    self.assertEquals(self.row.get('badkey', 'raargh'), 'raargh')

  def testUpdateKey(self):
    """ResultRow can have its values updated using dictionary assignment"""
    self.row['age'] = 28
    self.assertEquals(self.row['age'], 28)

  def testAddKey(self):
    """ResultRow can have field + value combinations added, like a dict"""
    self.row['role'] = 'developer'
    self.assertEquals(self.row['role'], 'developer')
    self.assertEquals(self.row[3], 'developer')
    self.assertEquals(self.row.popitem(), ('role', 'developer'))

  def testDeleteIndex(self):
    """ResultRow entries can be deleted using the index"""
    del self.row[1]
    self.assertEquals(self.row.items(), [('name', 'Elmer'), ('age', 24)])
    self.assertRaises(sqlresult.FieldError, self.row.__getitem__, 'nationality')

  def testDeleteIndexError(self):
    """ResultRow throws a FieldError when attempting to delete a bad index"""
    self.assertRaises(sqlresult.FieldError, self.row.__delitem__, 100)

  def testDeleteKey(self):
    """ResultRow entries can be deleted using the fieldname"""
    del self.row['nationality']
    self.assertEquals(self.row.items(), [('name', 'Elmer'), ('age', 24)])
    self.assertRaises(sqlresult.FieldError, self.row.__getitem__, 'nationality')

  def testDeleteKeyError(self):
    """ResultRow throws a FieldError when attempting to delete a bad key"""
    self.assertRaises(sqlresult.FieldError, self.row.__delitem__, 'badkey')


class ResultRowDataExtraction(unittest.TestCase):
  """Ensure that all name/value/item collecting methods work as intended."""
  def setUp(self):
    """Set up a persistent test environment."""
    self.names = ['name', 'project', 'age', 'language',
                  'character', 'distro-of-choice']
    self.values = ('Elmer', 'Underdark', 24, 'python', 'geek', u'Ubuntu')
    self.row = sqlresult.ResultRow(self.names, self.values)

  def testGetIterator(self):
    """ResultRow's iterator yields all the values stored."""
    self.assertEquals(tuple(self.row), self.values)

  def testGetFieldnames(self):
    """ResultRow's names property returns a list of all fieldnames stored."""
    self.assertEquals(self.row.names, self.names)

  def testGetItems(self):
    """ResultRow's item property return a list of all item pairs as tuples."""
    self.assertEquals(
        self.row.items(), zip(self.names, self.values))
    self.assertEquals(
        dict(self.row.items()), dict(zip(self.names, self.values)))


class ResultRowEquality(unittest.TestCase):
  """Ensure ResultRow compares properly to other instances and objects."""
  def setUp(self):
    """Set up a persistent test environment."""
    self.names = ['name', 'nationality', 'age']
    self.values = ('Elmer', 'Dutch', 24)
    self.row = sqlresult.ResultRow(self.names, self.values)

  def testEqualityWithResultRow(self):
    """ResultRow is equal to another ResultRow with the same contents."""
    other = sqlresult.ResultRow(self.names, self.values)
    self.assertNotEqual(id(self.row), id(other))
    self.assertEquals(self.row, other)

  def testNonEqualityWithNonResultRow(self):
    """ResultRow is not equal to a tuple or list with the same contents."""
    self.assertNotEqual(self.row, self.values)
    self.assertNotEqual(self.row, list(self.values))

  def testNonEqualityWithDifferentNames(self):
    """ResultRow is not equal to a ResultRow with different fieldnames."""
    names_copy = self.names[:]
    names_copy[1] = 'not the same'
    other_names = sqlresult.ResultRow(names_copy, self.values)
    self.assertNotEqual(self.row, other_names)

  def testNonEqualityWithDifferentValues(self):
    """ResultRow is not equal to a ResultRow with different values."""
    values_copy = list(self.values)
    values_copy[1] = 'Bob'
    other_values = sqlresult.ResultRow(self.names, values_copy)
    self.assertNotEqual(self.row, other_values)

  def testNonEqualityWithDifferentSize(self):
    """ResultRow is not equal to a ResultSet with a different length."""
    self.names.append('animal')
    values_copy = list(self.values)
    values_copy.append('pink ponies')
    bigger = sqlresult.ResultRow(self.names, values_copy)
    self.assertNotEqual(self.row, bigger)
    self.assertNotEqual(bigger, self.row)


class ResultSetBasicOperation(unittest.TestCase):
  """Ensure basic functionality and usability."""
  def setUp(self):
    """Set up a persistent test environment."""
    self.fields = ('first', 'second', 'third', 'fourth')
    self.result = tuple(tuple(2 ** i for i in range(j, j + 4))
                        for j in range(0, 13, 4))

  def testFalseWhenEmpty(self):
    """ResultSet is boolean False when there's all but a result."""
    result = sqlresult.ResultSet(query='Lorem Ipsum dolor sit amet.',
                                 charset='latin1',
                                 result=(),
                                 fields=self.fields,
                                 affected=1,
                                 insertid=1)
    self.assertFalse(result)

  def testTrueWhenFilled(self):
    """ResultSet is boolean True when it has a result stored."""
    result = sqlresult.ResultSet(query='',
                                 charset='',
                                 result=self.result,
                                 fields=self.fields,
                                 affected=0,
                                 insertid=0)
    self.assertTrue(result)

if __name__ == '__main__':
  unittest.main(testRunner=unittest.TextTestRunner(verbosity=2))



#!/usr/bin/python3
"""Module to send emails through an smtp server"""

__author__ = 'Elmer de Looff <elmer@underdark.nl>'
__version__ = '0.3'

# Standard modules
import base64
import os
import smtplib
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from uweb3.libs.safestring import  EmailAddresssafestring, EmailHeadersafestring


class MailError(Exception):
  """Something went wrong sending your email"""


class MailSender:
  """Easy context-interface for sending mail."""
  def __init__(self, host='localhost', port=25,
               local_hostname=None, timeout=5):
    """Sets up the connection to the SMTP server.

    Arguments:
      % host: str ~~ 'localhost'
        The SMTP hostname to connect to.
      % port: int ~~ 25
        Port for the SMTP server.
      % local_hostname: str ~~ from local hostname
        The hostname for which we want to send messages.
      % timeout: int ~~ 5
        Timeout in seconds.
    """
    self.server = None
    self.options = {'host': host, 'port': port,
                    'local_hostname': local_hostname or os.uname()[1],
                    'timeout': timeout}

  def __enter__(self):
    """Returns a SendMailContext for sending emails."""
    try:
      self.server = smtplib.SMTP(**self.options)
    except ConnectionRefusedError as error:
      raise SMTPConnectError(error, 'Connection refused.')
    return SendMailContext(self.server)

  def __exit__(self, *_exc_args):
    """Done sending mail, closes the smtp server connection."""
    self.server.quit()


class SendMailContext:
  """Context to use for sending emails."""
  def __init__(self, server):
    """Stores the server object locally."""
    self.server = server

  def Text(self, recipients, subject, content,
           sender=None, reply_to=None, charset='utf8'):
    """Send a text message

    Arguments:
      @ recipients: str / list of str
        Email address(es) of all TO: recipients.
      @ subject:  str
        Email subject
      @ content: str
        Body of the email
      % sender: str ~~ self.Noreply()
        The sender email addres, this defaults to the no-reply address.
      % reply_to: str ~~ None
        Optional reply-to address that differs from sender.
      % charset: str ~~ 'utf8'
        Character set to encode mail to.
    """
    message = MIMEMultipart()
    message['From'] = EmailAddresssafestring('') + (sender or self.Noreply())
    message['To'] = self.ParseRecipients(recipients)
    message['Subject'] = EmailHeadersafestring('') + ' '.join(subject.strip().split())
    message.attach(MIMEText(content.encode(charset), 'plain', charset))
    if reply_to:
      message['Reply-to'] = self.ParseRecipients(reply_to)
    self.server.sendmail(message['From'], recipients, message.as_string())

  def Attachments(self, recipients, subject, content,
                  attachments, sender=None, reply_to=None, charset='utf8'):
    """Sends email with attachments.

    Arguments like `Text()` but adds `attachments` after content. This should
    be a list of `str` (filename), `file` or 2-tuples with name and content.
    Content in case of 2-tuple can be `str` or any file-like object.
    """
    message = MIMEMultipart()
    message['From'] = EmailAddresssafestring('') + (sender or self.Noreply())
    message['To'] = self.ParseRecipients(recipients)
    message['Subject'] = EmailHeadersafestring('') + ' '.join(subject.strip().split())
    if reply_to:
      message['Reply-to'] = self.ParseRecipients(reply_to)
    message.attach(MIMEText(content.encode(charset), 'plain', charset))
    if isinstance(attachments, str):
      message.attach(self.ParseAttachment(attachments))
    else:
      for attachment in attachments:
        message.attach(self.ParseAttachment(attachment))
    self.server.sendmail(message['From'], recipients, str(message))

  @staticmethod
  def ParseAttachment(attachment):
    """Parses an attachment descriptor and returns a MIMEBase part for email."""
    if isinstance(attachment, tuple):
      name, contents = attachment
      if hasattr(contents, 'read'):
        contents = contents.read()
    elif isinstance(attachment, str):
      name = os.path.basename(attachment)
      contents = file(attachment, 'rb').read()
    elif isinstance(attachment, file):
      name = os.path.basename(attachment.name)
      attachment.seek(0)
      contents = attachment.read()

    part = MIMEBase('application', 'octet-stream')
    part.set_payload(Wrap(base64.b64encode(contents)))
    part.add_header('Content-Transfer-Encoding', 'base64')
    part.add_header('Content-Disposition', 'attachment; filename="%s"' % name)
    return part

  @staticmethod
  def ParseRecipients(recipients):
    """Ensures multiple recipients are returned as a safestring without
    newlines."""
    if isinstance(recipients, str):
      return EmailAddresssafestring('') + recipients
    return EmailAddresssafestring('') + ', '.join(recipients)

  def Noreply(self):
    """Returns the no-reply email address for the configured local hostname."""
    return 'no-reply <no-reply@%s>' % self.server.local_hostname


def Wrap(content, cols=76):
  """Wraps multipart mime content into 76 column lines for niceness."""
  lines = []
  while content:
    lines.append(content[:cols])
    content = content[cols:]
  return '\r\n'.join(lines)


SMTPConnectError = smtplib.SMTPConnectError
SMTPRecipientsRefused = smtplib.SMTPRecipientsRefused



#!/usr/bin/python3
"""This file contains the connector for Mongo."""
__author__ = 'Jan Klopper (janunderdark.nl)'
__version__ = 0.1

from . import Connector

class Mongo(Connector):
  """Adds MongoDB support to connection manager object."""

  def __init__(self,  config, options, request, debug=False):
    """Returns a MongoDB database connection."""
    self.debug = debug
    import pymongo
    self.options = options.get(self.Name(), {})
    try:
      self.connection = pymongo.connection.Connection(
         host=self.options.get('host', 'localhost'),
         port=self.options.get('port', 27017))
      if 'database' in self.options:
        self.connection = self.connection[self.options['database']]
    except Exception as e:
      raise ConnectionError('Connection to "%s" of type "%s" resulted in: %r' % (self.Name(), type(self), e))

  def Disconnect(self):
    """Closes the Mongo connection."""
    if self.debug:
      print('%s closed connection to: %r' % (self.Name(), self.options.get('database', 'Unspecified')))
    self.connection.close()
    del(self.connection)



#!/usr/bin/python3
"""uWeb3 PageMaker class and its various Mixins."""

import datetime
import logging
import mimetypes
import os
import pyclbr
import sys
import threading
import time
import hashlib
import glob
from base64 import b64encode
from pymysql import Error as pymysqlerr


import uweb3
from ..connections import ConnectionManager
from .. import response, templateparser

RFC_1123_DATE = '%a, %d %b %Y %T GMT'

class ReloadModules(Exception):
  """Signals the handler that it should reload the pageclass"""


class CacheStorage:
  """A (semi) persistent storage for the PageMaker."""
  def __init__(self):
    super(CacheStorage, self).__init__()
    self._dict = {}
    self._lock = threading.RLock()

  def __contains__(self, key):
    return key in self._dict

  def Del(self, key):
    """Removes the given key from the persistent storage.

    N.B. if the key was not in the persistent storage, no error is raised.
    """
    with self._lock:
      try:
        del self._dict[key]
      except KeyError:
        pass  # If a key is not in the storage, consider this a success

  def Get(self, key, *default):
    """Returns the current value for `key`, or the `default` if it doesn't."""
    with self._lock:
      if len(default) > 1:
        raise ValueError('Only one default value accepted')
      try:
        return self._dict[key]
      except KeyError:
        if default:
          return default[0]
        raise

  def Set(self, key, value):
    """Sets the `key` in the dictionary storage to `value`."""
    self._dict[key] = value

  def SetDefault(self, key, default=None):
    """Returns the value for `key` or sets it to `default` if it doesn't exist.

    Arguments:
      @ key: obj
        The key to retrieve from the dictionary storage.
      @ default: obj ~~ None
        The default new value for the given key if it doesn't exist yet.
    """
    with self._lock:
      return self._dict.setdefault(key, default)


class MimeTypeDict(dict):
  """Dictionary that defines special behavior for mimetypes.

  Mimetypes (of typical form "type/subtype") are stored as (type, subtype) keys.
  This allows grouping of types to happen, and fallbacks to occur.

  The following is a typical complete MIMEType example:
    >>> mime_type_dict['text/html'] = 'HTML content'

  One could also define a default for the whole type, as follows:
    >>> mime_type_dict['text/*'] = 'Default'Exception
    >>> mime_type_dict['text/nonexistant']
    'Default'
  """
  def __init__(self, data=(), **kwds):
    super(MimeTypeDict, self).__init__()
    if data:
      self.update(data)
    if kwds:
      self.update(**kwds)

  @staticmethod
  def MimeSplit(mime):
    """Split up a MIMEtype in a type and subtype, return as tuple.

    When the subtype if undefined or '*', only the type is returned, as 1-tuple.
    """
    mime_type, _sep, mime_subtype = mime.lower().partition('/')
    if not mime_subtype or mime_subtype == '*':
      return mime_type,  # 1-tuple
    return mime_type, mime_subtype

  def __setitem__(self, mime, value):
    super(MimeTypeDict, self).__setitem__(self.MimeSplit(mime), value)

  def __getitem__(self, mime):
    parsed_mime = self.MimeSplit(mime)
    try:
      return super(MimeTypeDict, self).__getitem__(parsed_mime)
    except KeyError:
      try:
        return super(MimeTypeDict, self).__getitem__(parsed_mime[:1])
      except KeyError:
        raise KeyError('KeyError: %r' % mime)

  def get(self, mime, default=None):
    try:
      return self[mime]
    except KeyError:
      return default

  def update(self, data=None, **kwargs):
    """Update the dictionary with new values from another dictionary.

    Also takes values from an iterable object of pairwise data.
    """
    if data:
      try:
        for key, value in data.items():
          self[key] = value
      except AttributeError:
        # Argument data is not a proper dict, treat it as an iterable of tuples.
        for key, value in data:
          self[key] = value
    if kwargs:
      self.update(kwargs)


class XSRFToken:
  def __init__(self, seed, remote_addr):
    self.seed = seed
    self.remote_addr = remote_addr
    self.unix_today = time.mktime(datetime.datetime.now().date().timetuple())

  def generate_token(self):
    """Generate an XSRF token

    XSRF token is generated based on the unix timestamp from today,
    a randomly generated seed and the IP addres from the request
    """
    hashed = (str(self.unix_today) + self.seed + self.remote_addr).encode('utf-8')
    h = hashlib.new('ripemd160')
    h.update(hashed)
    return h.hexdigest()


class Base:
  # Constant for persistent storage accross requests. This will be accessible
  # by all threads of the same application (in the same Python process).
  PERSISTENT = CacheStorage()
  # Base paths for templates and public data. These are used in the PageMaker
  # classmethods that set up paths specific for that pagemaker.
  TEMPLATE_DIR = 'templates'

  def __init__(self):
    self.persistent = self.PERSISTENT
    # clean up any request tags in the template parser, We do this in the init
    # because due to crashes we might not have triggered any __del__ or similar
    # end of request code.
    if '__parser' in self.persistent:
      self.persistent.Get('__parser').ClearRequestTags()

  def _PostInit(self):
    pass

  @property
  def parser(self):
    """Provides a templateparser.Parser instance.

    If the config file specificied a [templates] section and a `path` is
    assigned in there, this path will be used.
    Otherwise, the `TEMPLATE_DIR` will be used to load templates from.
    """
    if '__parser' not in self.persistent:
      self.persistent.Set('__parser', templateparser.Parser(
          self.options.get('templates', {}).get('path', self.TEMPLATE_DIR)))
    return self.persistent.Get('__parser')


class WebsocketPageMaker(Base):
  """Pagemaker for the websocket routes.
  This is different from the BasePageMaker as we choose to not have a database connection
  in our WebSocketPageMaker.

  This class lacks pretty much all functionality that the BasePageMaker has.
  """

  #TODO: Add request to pagemaker?
  def Connect(self, sid, env):
    """This is the connect event,
    sets the req variable that contains the request headers.
    """
    print(f"User connected with SocketID {sid}: ")
    self.req = env


class XSRFMixin:
  """Provides XSRF protection by enabling setting xsrf token cookies, checking
  them and setting a flag based on their value

  A seperate decorator can then be used to clear the POST/GET/PUT variables if
  needed in specific pagemaker functions depending on that page's security
  context.
  """
  XSRFCOOKIE = 'xsrf'
  XSRF_seed = str(os.urandom(32))

  def validatexsrf(self):
    """Sets the invalid_xsrf_token flag to true or false"""
    self.invalid_xsrf_token = False
    if self.req.method != 'GET': # GET calls will be ignored, but will set a cookie
      self.invalid_xsrf_token = True
      try:
        user_supplied_xsrf_token = getattr(self, self.req.method.lower()).getfirst(self.XSRFCOOKIE)
        self.invalid_xsrf_token = (self.cookies.get(self.XSRFCOOKIE) != user_supplied_xsrf_token)
      except Exception:
        # any error in looking up the cookie of the supplied post vars will result in a invalid xsrf token flag
        pass
    # If no cookie is present, set it.
    self._Set_XSRF_cookie()

  def _Set_XSRF_cookie(self):
    """This creates a new XSRF token for this client, which is IP bound, and
    stores it in a cookie.
    """
    xsrf_cookie = self.cookies.get(self.XSRFCOOKIE, False)
    if not xsrf_cookie:
      xsrf_cookie = XSRFToken(self.XSRF_seed, self.req.env['REAL_REMOTE_ADDR']).generate_token()
      self.req.AddCookie(self.XSRFCOOKIE, xsrf_cookie, path="/", httponly=True)
    return xsrf_cookie

  def XSRFInvalidToken(self):
    """Returns an error message regarding an incorrect XSRF token."""
    errorpage = templateparser.FileTemplate(os.path.join(
        os.path.dirname(__file__), 'http_403.html'))
    error = """Your browser did not send us the correct token, any token at all, or a timed out token.
    Because of this we cannot allow you to perform this action at this time. Please refresh the previous page and try again."""

    return uweb3.Response(content=errorpage.Parse(error=error),
        httpcode=403, headers=self.req.response.headers)

  def _Get_XSRF(self):
    """Easy access to the XSRF token"""
    try:
      return self.cookies[self.XSRFCOOKIE]
    except KeyError:
      return self._Set_XSRF_cookie()


class LoginMixin:
  """This mixin provides a few methods that help with handling logins, sessions
  and related database/cookie interaction"""

  def _ReadSession(self):
    return NotImplemented

  @property
  def user(self):
    """Returns the current user"""
    if not hasattr(self, '_user') or not self._user:
      try:
        self._user = self._ReadSession()
      except ValueError:
        self._user = False
    return self._user


class BasePageMaker(Base):
  """Provides the base pagemaker methods for all the html generators."""
  _registery = []

  # Default Static() handler cache durations, per MIMEtype, in days
  PUBLIC_DIR = 'static'
  CACHE_DURATION = MimeTypeDict({'text': 7, 'image': 30, 'application': 7,
      'text/css': 7})

  def __init__(self,
              req,
              config=None,
              executing_path=None):
    """sets up the template parser and database connections.

    Arguments:
      @ req: request.Request
        The originating request, including environment, GET, POST and cookies.
      % config: dict ~~ None
        Configuration for the pagemaker, with database connection information
        and other settings. This will be available through `self.options`.
      % executing_path: str/path
        This is the path to the uWeb3 routing file.
    """
    super(BasePageMaker, self).__init__()
    self.__SetupPaths(executing_path)
    self.req = req
    self.cookies = req.vars['cookie']
    self.get = req.vars['get']
    self.post = req.vars['post'] if 'post' in req.vars else {}
    self.put = req.vars['put'] if 'put' in req.vars else {}
    self.delete = req.vars['delete'] if 'delete' in req.vars else {}
    self.config = config or None
    self.options = config.options if config else {}
    self.debug = DebuggerMixin in self.__class__.__mro__
    try:
      self.connection = self.persistent.Get('connection')
    except KeyError:
      self.persistent.Set('connection', ConnectionManager(self.config, self.options, self.debug))
      self.connection = self.persistent.Get('connection')

  @classmethod
  def LoadModules(cls, routes='routes/*.py'):
    """Loops over all .py files apart from some exceptions in target directory
    Looks for classes that contain pagemaker

    Arguments:
      % default_routes: str
        Location to your route files. Defaults to routes/*.py
        Supports glob style syntax, non recursive.
    """
    bases = []
    for file in glob.glob(routes):
      module = os.path.relpath(os.path.join(os.getcwd(), file[:-3])).replace('/', '.')
      classlist = pyclbr.readmodule_ex(module)
      for name, data in classlist.items():
        if hasattr(data, 'super') and 'PageMaker' in data.super[0]:
          module = __import__(file, fromlist=[name])
          bases.append(getattr(module, name))
    return bases

  def _PostInit(self):
    """Method that gets called for derived classes of BasePageMaker."""

  def _ConnectionRollback(self):
    """Roll back all connections, this method can be overwritten by the user"""
    self.connection.Rollback()

  @classmethod
  def __SetupPaths(cls, executing_path):
    """This sets up the correct paths for the PageMaker subclasses.

    From the passed in `cls`, it retrieves the filename. Of that path, the
    directory is used as the working directory. Then, the module constant
    TEMPLATE_DIR is used to define class constants from.
    """
    local_file = os.path.abspath(sys.modules[cls.__module__].__file__)
    cls.LOCAL_DIR = cls_dir = executing_path
    cls.PUBLIC_DIR = os.path.realpath(os.path.join(cls_dir, cls.PUBLIC_DIR))
    cls.TEMPLATE_DIR = os.path.realpath(os.path.join(cls_dir, cls.TEMPLATE_DIR))

  def Static(self, rel_path):
    """Provides a handler for static content.

    The requested `path` is truncated against a root (removing any uplevels),
    and then added to the working dir + PUBLIC_DIR. If the request file exists,
    then the requested file is retrieved, its mimetype guessed, and returned
    to the client performing the request.

    Should the requested file not exist, a 404 page is returned instead.

    Arguments:
      @ rel_path: str
        The filename relative to the working directory of the webserver.

    Returns:
      Page: contains the content and mimetype of the requested file, or a 404
            page if the file was not available on the local path.
    """

    abs_path = os.path.realpath(os.path.join(self.PUBLIC_DIR, rel_path.lstrip('/')))
    print(abs_path, self.PUBLIC_DIR)

    if os.path.commonprefix((abs_path, self.PUBLIC_DIR)) != self.PUBLIC_DIR:
      return self._StaticNotFound(rel_path)
    try:
      content_type, _encoding = mimetypes.guess_type(abs_path)
      if not content_type:
        content_type = 'text/plain'
      mtime = os.path.getmtime(abs_path)
      length = os.path.getsize(abs_path)
      with open(abs_path, 'rb') as staticfile:
        cache_days = self.CACHE_DURATION.get(content_type, 0)
        expires = datetime.datetime.utcnow() + datetime.timedelta(cache_days)
        return response.Response(content=staticfile.read(),
                        content_type=content_type,
                        headers={'Expires': expires.strftime(RFC_1123_DATE),
                                 'cache-control': 'max-age=%d' %
                                    (cache_days*24*60*60),
                                 'last-modified': time.ctime(mtime),
                                 'content-length': length})
    except IOError:
      return self._StaticNotFound(rel_path)

  def _StaticNotFound(self, _path):
    message = 'This is not the path you\'re looking for. No such file %r' % (
      self.req.env['PATH_INFO'])
    return response.Response(message, content_type='text/plain', httpcode=404)

  def _NotFound(self, _path):
    message = 'This is not the path you\'re looking for. No such path %r' % (
      self.req.env['PATH_INFO'])
    return response.Response(message, content_type='text/html', httpcode=404)

  def InternalServerError(self, exc_type, exc_value, traceback):
    """Returns a plain text notification about an internal server error."""
    error = 'INTERNAL SERVER ERROR (HTTP 500) DURING PROCESSING OF %r' % (
                self.req.path)
    self.req.registry.logger.error(
        error, exc_info=(exc_type, exc_value, traceback))
    return response.Response(
        content=error, content_type='text/plain', httpcode=500)

  @staticmethod
  def Reload():
    """Raises `ReloadModules`, telling the Handler() to reload its pageclass."""
    raise ReloadModules('Reloading ... ')

  def _PostRequest(self):
    """Method that gets called after each request"""
    self.connection.PostRequest()


class DebuggerMixin:
  """Replaces the default handler for Internal Server Errors.

  This one prints a host of debugging and request information, though it still
  lacks interactive functions.
  """
  CACHE_DURATION = MimeTypeDict({})
  ERROR_TEMPLATE = 'http_500.html'

  def _ParseStackFrames(self, stack):
    """Generates list items for traceback information.

    Each traceback item contains the file- and function name, the line numer
    and the source that belongs with it. For each stack frame, the local
    variables are also added to it, allowing proper analysis to happen.

    This most likely doesn't need overriding / redefining in a subclass.

    Arguments:
      @ stack: traceback.stack
        The stack frames to return analysis on.

    Yields:
      str: Template-parsed HTML with frame information.
    """
    frames = []
    while stack:
      frame = stack.tb_frame
      frames.append({'file': os.path.abspath(frame.f_code.co_filename),
                     'scope': frame.f_code.co_name,
                     'locals': frame.f_locals,
                     'source': self._SourceLines(
                         frame.f_code.co_filename, frame.f_lineno)})
      stack = stack.tb_next
    return reversed(frames)

  @staticmethod
  def _SourceLines(filename, line_num, context=3):
    """Yields the offending source line, and `context` lines of context.

    Arguments:
      @ filename: str
        The filename (path) of which we should print lines of source.
      @ line_num: int
        The line number for the offending line.
      % context: int ~~ 3
        Number of lines context, before and after the offending line.

    Yields:
      str: Templated list-item for a source code line.
    """
    import linecache
    for line_num in range(line_num - context, line_num + context + 1):
      yield line_num, linecache.getline(filename, line_num)

  def InternalServerError(self, exc_type, exc_value, traceback):
    """Returns a HTTP 500 response with detailed failure analysis."""
    self.req.registry.logger.error(
        'INTERNAL SERVER ERROR (HTTP 500) DURING PROCESSING OF %r',
        self.req.path, exc_info=(exc_type, exc_value, traceback))
    exception_data = {
        'cookies': self.cookies,
        'environ': self.req.env,
        'query_args': self.get,
        'post_data': self.post,
        'error_for_error': False,
        'exc': {'type': exc_type, 'value': exc_value,
                'traceback': self._ParseStackFrames(traceback)}}

    error_template = templateparser.FileTemplate(os.path.join(
      os.path.dirname(__file__), self.ERROR_TEMPLATE))
    try:
      return response.Response(
          error_template.Parse(**exception_data), httpcode=500)
    except Exception:
      exc_type, exc_value, traceback = sys.exc_info()
      self.req.registry.logger.critical(
          'INTERNAL SERVER ERROR (HTTP 500) DURING PROCESSING OF ERROR PAGE',
          exc_info=(exc_type, exc_value, traceback))
      exception_data['error_for_error'] = True
      exception_data['orig_exc'] = exception_data['exc']
      exception_data['exc'] = {'type': exc_type, 'value': exc_value,
                               'traceback': self._ParseStackFrames(traceback)}
      return response.Response(
          error_template.Parse(**exception_data), httpcode=500)


class CSPMixin:
  """Provides CSP header output.

  https://content-security-policy.com/
  """
  _csp = {
        "default-src": ("'none'",),
        "object-src": ("'none'",),
        "script-src":  ("'none'",),
        "style-src":  ("'none'",),
        "form-action":  ("'none'",),
        "connect-src":  ("'none'",),
        "img-src":  ("'none'",),
        "font-src":  ("'none'",),
        "frame-ancestors":  ("'none'",),
        "base-uri": ("'none'",)
  }

  def _SetCsp(self, resourcetype="default-src", urls=("'self'", ), append=True):
    """Add a new CSP url to the csp headers for the given resourcetype.

    resourcetype is any of the CSP resource types as defined in:
      https://content-security-policy.com/#directive
      defaults to: default-src

    urls should be one or more of:
      https://content-security-policy.com/#source_list
      default to 'self'
      string or tuple/list is allowed

    By default this appends to the already present list of sources for the given
    resourcetype

    """
    if isinstance(urls, str):
      urls = [urls, ]
    else:
      urls = list(urls) if type(urls) == tuple else urls
    if resourcetype not in self._csp:
      self._csp[resourcetype] = []
    if self._csp[resourcetype] == "'none'" or not append:
      self._csp[resourcetype] = urls
      return
    self._csp[resourcetype].extend(urls)

  def _CSPFromConfig(self, config):
    """sets the CSP headers from a Dictionary
    Dict keys should be resourcetypes, values should be lists of urls

    resourcetype is any of the CSP resource types as defined in:
      https://content-security-policy.com/#directive

    urls are in the form:
      https://content-security-policy.com/#source_list
    """
    self._csp = config

  def _CSPheaders(self):
    """Adds the constructed CSP header to the request"""
    csp = '; '.join(
        "%s %s" % (key, ' '.join(value)) for key, value in self._csp.items())
    self.req.AddHeader('Content-Security-Policy', csp)

# ##############################################################################
# Classes for public use (wildcard import)
#
class PageMaker(XSRFMixin, BasePageMaker):
  """The basic PageMaker class, providing XSRF support."""


class DebuggingPageMaker(DebuggerMixin, PageMaker):
  """The same basic PageMaker, with added debugging on HTTP 500."""



#!/usr/bin/python3
"""This file contains the connector for Sqlite."""
__author__ = 'Jan Klopper (janunderdark.nl)'
__version__ = 0.1

from . import Connector

class Sqlite(Connector):
  """Adds SQLite support to connection manager object."""

  def __init__(self,  config, options, request, debug=False):
    """Returns a SQLite database connection.
    The name of the class is used as the local filename.
    """
    from ..libs.sqltalk import sqlite
    self.debug = debug
    self.options = options[self.Name()]
    try:
      self.connection = sqlite.Connect(self.options.get('database'))
    except Exception as e:
      raise ConnectionError('Connection to "%s" of type "%s" resulted in: %r' % (self.Name(), type(self), e))

  def Rollback(self):
    """Rolls back any uncommited transactions."""
    return self.connection.rollback()

  def Disconnect(self):
    """Closes the SQLite connection."""
    if self.debug:
      print('%s closed connection to: %r' % (self.Name(), self.options.get('database')))
    self.connection.close()
    del(self.connection)



#!/usr/bin/python3
"""Tests for the safestring code"""
__author__ = 'Jan Klopper (jan@underdark.nl)'
__version__ = 0.1

# default modules
import unittest

#custom modules
from uweb3.libs.safestring import URLsafestring, SQLSAFE, HTMLsafestring, URLqueryargumentsafestring, JSONsafestring, EmailAddresssafestring, Basesafestring

class BasesafestringMethods(unittest.TestCase):
  def test_creation_str(self):
    with self.assertRaises(NotImplementedError) as context:
      str(Basesafestring('foo'))
    self.assertEqual(NotImplementedError, type(context.exception))

  def test_creation_repr(self):
    with self.assertRaises(NotImplementedError) as context:
      repr(Basesafestring('foo'))
    self.assertEqual(NotImplementedError, type(context.exception))


class TestHTMLStringMethods(unittest.TestCase):
  def test_init(self):
    testdata = HTMLsafestring('foo<test>')
    self.assertEqual(testdata, 'foo<test>')

  def test_init_unsafe(self):
    testdata = HTMLsafestring('foo<test>', unsafe=True)
    self.assertEqual(testdata, 'foo&lt;test&gt;')

  def test_addition(self):
    testdata = HTMLsafestring('foo') + '<b>test'
    self.assertEqual(testdata, 'foo&lt;b&gt;test')

  def test_addition_quotes(self):
    testdata = HTMLsafestring('foo') + '"test"'
    self.assertEqual(testdata, 'foo&quot;test&quot;')

  def test_addition_same(self):
    testdata = HTMLsafestring('foo<test>') + HTMLsafestring('foo<test>')
    self.assertEqual(testdata, 'foo<test>foo<test>')

  def test_addition_same_escaped(self):
    testdata = HTMLsafestring('foo') + '<test>'
    testdata = testdata + HTMLsafestring('foo<test>')
    self.assertEqual(testdata, 'foo&lt;test&gt;foo<test>')

  def test_addition_other(self):
    testdata = HTMLsafestring('foo<test>')
    urlquery = URLqueryargumentsafestring('foo') + ' test test/test'
    self.assertEqual(testdata + urlquery, 'foo<test>foo test test/test')

  def test_format(self):
    testdata = HTMLsafestring('foo<test> {} test').format('<b>')
    self.assertEqual(testdata, 'foo<test> &lt;b&gt; test')

  def test_format_nested_safesametype(self):
    testdata = HTMLsafestring('test {} {}').format(
      '<b>', HTMLsafestring('<b>'))
    self.assertEqual(testdata, 'test &lt;b&gt; <b>')

  def test_format_nested_safeothertype(self):
    testdata = HTMLsafestring('test {} {}').format(
      '<b>', EmailAddresssafestring('<b>'))
    self.assertEqual(testdata, 'test &lt;b&gt; &lt;b&gt;')

  def test_format_keyword(self):
    testdata = HTMLsafestring('foo<test> {kw} test').format(kw='<b>')
    self.assertEqual(testdata, 'foo<test> &lt;b&gt; test')

  def test_join(self):
    """Tests to join two already safe list items"""
    testdata = HTMLsafestring('').join((HTMLsafestring('<b>'),
                                        HTMLsafestring('<b>')))
    self.assertEqual(testdata, '<b><b>')
    self.assertIsInstance(testdata, HTMLsafestring)

  def test_join_unsafe(self):
    """Test a join over possibly insafe and safe strings combined"""
    testdata = HTMLsafestring('').join(('<b>',
                                        HTMLsafestring('<b>')))
    self.assertEqual(testdata, '&lt;b&gt;<b>')
    self.assertIsInstance(testdata, HTMLsafestring)


class TestJSonStringMethods(unittest.TestCase):
  def test_addition(self):
    testdata = JSONsafestring('foo') + '"test"'
    self.assertEqual(testdata, 'foo"\\"test\\""')

class TestURLqueryargumentsafestringMethods(unittest.TestCase):
  def test_addition(self):
    testdata = URLqueryargumentsafestring('foo') + 'test test/test'
    self.assertEqual(testdata, 'footest+test%2Ftest')


class TestURLsafestringMethods(unittest.TestCase):
  def test_headerinjection(self):
    """See if we correctly clean up header injection attemps"""
    testdata = URLsafestring('https://underdark.nl/somefile\n\nlocation: http://attacker.nl', unsafe=True)
    self.assertEqual(str(testdata), 'https://underdark.nl/somefile')

  def test_validurlunwisechars(self):
    url = 'http://mw1.google.com/mw-earth-vectordb/kml-samples/gp/seattle/gigapxl/$[level]/r$[y]_c$[x].jpg#fragment'
    testdata = URLsafestring(url, unsafe=True)
    self.assertEqual(str(testdata), url)

class TestEmailAddresssafestringMethods(unittest.TestCase):
  def test_unsafe_init(self):
    """See if we correctly clean up header injection attemps"""
    testdata = EmailAddresssafestring('jan@underdark.nl\nbcc: victim@otherhost.com', unsafe=True)
    self.assertEqual(testdata, 'jan@underdark.nl')

class TestSQLSAFEMethods(unittest.TestCase):
  def test_user_supplied_safe_value(self):
    user_supplied_safe_object = SQLSAFE("SELECT * FROM users WHERE username = 'username\t'")
    self.assertEqual(user_supplied_safe_object, "SELECT * FROM users WHERE username = 'username\t'")
    self.assertIsInstance(user_supplied_safe_object, SQLSAFE)

  def test_escaping_wrong_values_type(self):
    with self.assertRaises(ValueError):
      self.assertRaises(SQLSAFE("""SELECT * FROM users WHERE username = ?""", values=["username'"], unsafe=True))

  def test_escaping_uneven_replacements_and_values(self):
    with self.assertRaises(ValueError):
      self.assertRaises(SQLSAFE("""SELECT * FROM users WHERE username = ?""", values=["username'", "test"], unsafe=True))
      self.assertRaises(SQLSAFE("""SELECT * FROM users WHERE username = ? AND name=?""", values=["username'"], unsafe=True))

  def test_escaping(self):
    testdata = SQLSAFE("""SELECT * FROM users WHERE username = ?""", values=("username'",), unsafe=True)
    self.assertEqual(testdata, "SELECT * FROM users WHERE username = 'username\\''")
    testdata = SQLSAFE("""SELECT * FROM users WHERE username = ?""", values=('username"',), unsafe=True)
    self.assertEqual(testdata, "SELECT * FROM users WHERE username = 'username\\\"'")
    testdata = SQLSAFE("""SELECT * FROM users WHERE username = ? AND ? """, values=('username"', "password"), unsafe=True)

  def test_concatenation(self):
    testdata = SQLSAFE("""SELECT * FROM users WHERE username = ?""", values=("username'",), unsafe=True)
    other = "AND firstname='test'"
    self.assertEqual(testdata + other, "SELECT * FROM users WHERE username = 'username\\'' AND firstname=\\'test\\'")
    testdata = SQLSAFE("""SELECT * FROM users WHERE username = ?""", values=('username"',), unsafe=True)
    other = "AND firstname='test'"
    self.assertEqual(testdata + other, "SELECT * FROM users WHERE username = 'username\\\"' AND firstname=\\'test\\'")

  def test_unescape_wrong_type(self):
    """Validate if the string we are trying to unescape is part of an SQLSAFE instance"""
    testdata = SQLSAFE("""SELECT * FROM users WHERE username = ?""", values=("username'",), unsafe=True)
    with self.assertRaises(ValueError):
      self.assertRaises(testdata.unescape('whatever'))

  def test_correct_escape_character(self):
    """Validate that all characters are escaped as expected"""
    self.assertEqual(SQLSAFE.sanitize('\0', with_quotes=False), '\\0')
    self.assertEqual(SQLSAFE.sanitize('\b', with_quotes=False), '\\b')
    self.assertEqual(SQLSAFE.sanitize('\t', with_quotes=False), '\\t')
    self.assertEqual(SQLSAFE.sanitize('\n', with_quotes=False), '\\n')
    self.assertEqual(SQLSAFE.sanitize('\r', with_quotes=False), '\\r')
    self.assertEqual(SQLSAFE.sanitize('\x1a', with_quotes=False), '\\Z')
    self.assertEqual(SQLSAFE.sanitize('"', with_quotes=False), '\\"')
    self.assertEqual(SQLSAFE.sanitize('\'', with_quotes=False), '\\\'')
    self.assertEqual(SQLSAFE.sanitize('\\', with_quotes=False), '\\\\')

  def test_unescape(self):
    """Validate that the string is converted back to the original after escaping and unescaping"""
    testdata = SQLSAFE("""SELECT * FROM users WHERE username = ?""", values=("username\t \t",), unsafe=True)
    self.assertEqual(testdata, "SELECT * FROM users WHERE username = 'username\\t \\t'")
    self.assertEqual(testdata.unescape(testdata), "SELECT * FROM users WHERE username = 'username\t \t'")


if __name__ == '__main__':
    unittest.main()
